;-----------------------------------------------------------------------------
; PIC hardware UART test program, interrupt driven with circular buffers
;
; Copyright 1997 Eric Smith
;
; Permission is granted to use this code or portions thereof for
; non-commercial purposes provided the copyright notice is preserved intact
; and credit is given to the author.  For any other use, a license may be
; obtained from the author; send email to eric@brouhaha.com.
;
; $Id: uarttest.asm,v 1.3 1997/09/09 01:23:39 eric Exp eric $
;-----------------------------------------------------------------------------


	processor	16c65
	include		"p16c65.inc"
	radix		dec

	errorlevel	-224
	errorlevel	-305

	__config	_CP_OFF&_PWRTE_ON&_WDT_OFF&_HS_OSC


rambase		equ	20h

cmdtbl_page	equ	6
message_page	equ	7

xtalfreq	equ	4915200		; in Hz

rx_ring_size	equ	32
tx_ring_size	equ	32


;-----------------------------------------------------------------------------
; buad rate contants - (xtalfreq / (64 * baud)) - 1
;-----------------------------------------------------------------------------

b384	equ	(xtalfreq/(64*38400))-1
b192	equ	(xtalfreq/(64*19200))-1
b96	equ	(xtalfreq/(64*9600))-1
b48	equ	(xtalfreq/(64*4800))-1


;-----------------------------------------------------------------------------
; ascii characters
;-----------------------------------------------------------------------------

bel	equ	07h		; bell
bs	equ	08h		; back space
tab	equ	09h		; horizontal tab
lf	equ	0ah		; line feed
ff	equ	0ch		; form feed
cr	equ	0dh		; carriage return
xon	equ	11h		; transmit on
xoff	equ	13h		; transmit off
eof	equ	1ah		; end of file
esc	equ	1bh		; escape
sp	equ	20h		; space


;-----------------------------------------------------------------------------
; ring buffer macro
;-----------------------------------------------------------------------------

; advance buffer pointer, wrapping if necessary

; If the buffer size is a power of two, and the buffer is aligned on a
; multiple of twice its size, ringadv generates two instructions.  Otherwise
; it generates six instructions.

; further optimizations are possible for other special cases.

ringadv	macro	ptr,base,size
	local	pow2,aligned,bit,val
pow2	set	!(size&(size-1))
aligned	set	pow2&&((base&(size-1))==0)

	if	aligned
val	set	size
bit	set	0
	while	val>1
bit	set	bit+1
val	set	val>>1
	endw
	endif

	incf	ptr

	if	aligned&&!(base&(1<<bit))
	bcf	ptr,bit
	else
	movf	ptr,W
	xorlw	base+size
	movlw	base
	btfsc	STATUS,Z
	movwf	ptr
	endif
	endm


;-----------------------------------------------------------------------------
; ram
;-----------------------------------------------------------------------------

		org	rambase

; interrupt state save
irqw:		res	1		; must be in same loc in both banks

ramb2		equ	$+80h

irqstat:	res	1
irqfsr:		res	1


rx_ring_ip:	res	1
rx_ring_op:	res	1
rx_ring_cnt:	res	1

tx_ring_ip:	res	1
tx_ring_op:	res	1
tx_ring_cnt:	res	1

tx_temp:	res	1


char:		res	1
temp:		res	1
cmdptr:		res	1

cmd:		res	1

cmdflag:	res	1
gotarg1		equ	0
gotarg2		equ	1

arg1:		res	2
arg2:		res	2


		org	ramb2

rx_ring:	res	rx_ring_size
tx_ring:	res	tx_ring_size


;-----------------------------------------------------------------------------
; reset vector
;-----------------------------------------------------------------------------

	org	0

	goto	reset
	nop
	nop
	nop


;-----------------------------------------------------------------------------
; interrupt vector
;-----------------------------------------------------------------------------

	org	4

	movwf	irqw
	swapf	irqw
	swapf	STATUS,W
	bcf	STATUS,RP0
	movwf	irqstat
	movf	FSR,W
	movwf	irqfsr

	btfsc	PIR1,RCIF
	call	rxint

	btfsc	PIR1,TXIF
	call	txint


notx:	movf	irqfsr,W
	movwf	FSR
	swapf	irqstat,W
	movwf	STATUS
	swapf	irqw,W
	retfie


;-----------------------------------------------------------------------------
; UART receive interrupt
;-----------------------------------------------------------------------------

rxint:	movf	rx_ring_cnt,W	; is buffer full?
	xorlw	rx_ring_size
	btfsc	STATUS,Z
	goto	rx_overrun

	movf	rx_ring_ip,W	; store character in buffer
	movwf	FSR
	movf	RCREG,W
	movwf	INDF

	incf	rx_ring_cnt	; increment count

	ringadv	rx_ring_ip,rx_ring,rx_ring_size		; advance pointer

	return

rx_overrun:
	movf	RCREG,W		; read and discard character
	return


;-----------------------------------------------------------------------------
; UART receive interrupt
;-----------------------------------------------------------------------------

txint:	movf	tx_ring_cnt	; is buffer empty?
	btfsc	STATUS,Z
	goto	tx_underrun

	movf	tx_ring_op,W	; transmit char
	movwf	FSR
	movf	INDF,W
	movwf	TXREG

	decf	tx_ring_cnt	; decrement count

	ringadv	tx_ring_op,tx_ring,tx_ring_size		; advance pointer

	return

tx_underrun:
	movlw	PIE1
	movwf	FSR
	bcf	INDF,TXIE	; disable transmitter interrupt
	
	return


;-----------------------------------------------------------------------------
; UART initialization
;-----------------------------------------------------------------------------

uartinit:
	clrf	rx_ring_cnt
	movlw	rx_ring
	movwf	rx_ring_ip
	movwf	rx_ring_op

	clrf	tx_ring_cnt
	movlw	tx_ring
	movwf	tx_ring_ip
	movwf	tx_ring_op

	movlw	SPBRG		; set buad rate
	movwf	FSR
	movlw	b96
	movwf	INDF

	movlw	TXSTA		; enable transmitter
	movwf	FSR
	movlw	1<<TXEN
	movwf	INDF

	movlw	(1<<SPEN)|(1<<CREN)	; enable receiver
	movwf	RCSTA

	movlw	PIE1
	movwf	FSR
	bsf	INDF,RCIE	; enable receiver interrupt
	bsf	INTCON,PEIE

	return


;-----------------------------------------------------------------------------
; output a byte in hexadecimal
; uses temp
;-----------------------------------------------------------------------------

outhb:	movwf	temp
outhb1:	swapf	temp,w
	call	outhd
	movf	temp,w

; fall into outhd

;-----------------------------------------------------------------------------
; output a hex digit
;-----------------------------------------------------------------------------

outhd:	andlw	0fh
	addlw	0f6h
	btfsc	STATUS,C
	addlw	07h
	addlw	3ah

; fall into xmit

;-----------------------------------------------------------------------------
; serial transmit character
; blocks until space if available in buffer
;-----------------------------------------------------------------------------

xmit:	movwf	tx_temp

xmit0:	movf	tx_ring_cnt,W	; is there room for a character?
	xorlw	tx_ring_size
	btfsc	STATUS,Z
	goto	xmit0		; no, wait 'til there is

	movf	tx_ring_ip,W	; store char in buffer
	movwf	FSR
	movf	tx_temp,W
	movwf	INDF

	ringadv	tx_ring_ip,tx_ring,tx_ring_size	; advance pointer

	incf	tx_ring_cnt	; decrement tx char count
				; (this must not be done until after character
				; has been stored in buffer)

	movlw	PIE1
	movwf	FSR
	bsf	INDF,TXIE	; enable transmitter interrupt

	return


;-----------------------------------------------------------------------------
; serial receive character
; blocks until a character is available
;-----------------------------------------------------------------------------

recv:	movf	rx_ring_cnt	; is there anything in buffer?
	btfsc	STATUS,Z
	goto	recv		; no, wait 'til there is

	movf	rx_ring_op,w
	movwf	FSR

	ringadv	rx_ring_op,rx_ring,rx_ring_size	; advance pointer

	movf	INDF,W		; get character from buffer

	decf	rx_ring_cnt	; decrement rx char count
				; (this must not be done until after character
				; has been pulled from buffer)

	return


;-----------------------------------------------------------------------------
; message output subroutine
;-----------------------------------------------------------------------------

msgout:	movwf	temp
	movlw	message_page
	movwf	PCLATH
msglp:	movf	temp,W
	call	table
	xorlw	0
	btfsc	STATUS,Z
	return
	call	xmit
	incf	temp
	goto	msglp


;-----------------------------------------------------------------------------
; program entry point
;-----------------------------------------------------------------------------

reset:
	clrwdt
	movlw	b'10001001'	;Bit 0,1,2 = Prescaler divider = 4
				;Bit 3 = Prescaler assign: 0 for timer 0, 1 for WDT
				;Bit 4 = RTCC edge: 0 = low to high
				;Bit 5 = RTCC source: 0 = internal
				;Bit 6 = INT edge: 0 = interrupt on high to low
				;Bit 7 = Port B pull-ups: 1 = disabled
	option

	movlw	00h		; Port A - all pins unused, output 0
	movwf	PORTA
	movlw	000h
	tris	PORTA

	movlw	00h		; Port B - all pins unused, output 0
	movwf	PORTB
	movlw	00h
	tris	PORTB

; Port C
;   Bit 7 = UART receive input
;   Bit 6 = UART transmit output
;   Bits 5..0 unused, output 0

	movlw	00h		
	movwf	PORTC
	movlw	080h
	tris	PORTC

	movlw	00h		; Port D - all pins unused, output 0
	movwf	PORTD
	movlw	TRISD
	movwf	FSR
	movlw	00h
	movwf	INDF

	movlw	00h		; Port E - all pins unused, output 0
	movwf	PORTE
	movlw	TRISE
	movwf	FSR
	movlw	00h
	movwf	INDF


	clrf	INTCON

	call	uartinit

	bsf	INTCON,GIE	; now ready to handle interrupts


;-----------------------------------------------------------------------------
; main loop
;-----------------------------------------------------------------------------

main:	movlw	banner
	call	msgout

loop0:	clrf	arg1
	clrf	arg1+1
	clrf	arg2
	clrf	arg2+1
	bcf	cmdflag,gotarg1
	bcf	cmdflag,gotarg2

	movlw	prompt
	call	msgout

loop:	call	recv		; get and echo character
	call	toupper
	movwf	cmd		

	movf	cmd,w
	call	ishex
	btfss	STATUS,Z
	goto	nothex

	bsf	cmdflag,gotarg1

	movwf	temp		; rotate hex digit into arg1
	rlf	temp
	rlf	temp
	rlf	temp
	rlf	temp

	rlf	temp
	rlf	arg1
	rlf	arg1+1

	rlf	temp
	rlf	arg1
	rlf	arg1+1

	rlf	temp
	rlf	arg1
	rlf	arg1+1

	rlf	temp
	rlf	arg1
	rlf	arg1+1

	movf	cmd,w		; echo hex digit
	call	xmit	

	goto	loop


nothex:	movlw	cmdtbl&0xff
	movwf	cmdptr
	movf	cmd,w
	call	CmdScan

	btfss	STATUS,Z
	goto	loop
	goto	loop0


;-----------------------------------------------------------------------------
; tolower - if W contains 'A'..'Z', tranlate to 'a'..'z'
;-----------------------------------------------------------------------------

				; A-Z       a-z       0-9      
				; $41-$5a   $61-$7a   $30-$39   

tolower:
	addlw	0a5h		; $e6-$ff   $06-$1f   $d5-$de
	addlw	01ah		; $00-$19   $20-$39   $ef-$f8
	btfsc	STATUS,C
	addlw	020h		; $20-$39
	addlw	041h		; $61-$7a   $61-$7a   $30-$39
	return


;-----------------------------------------------------------------------------
; toupper - if W contains 'a'..'z', tranlate to 'A'..'Z'
;-----------------------------------------------------------------------------

				; A-Z       a-z       0-9      
				; $41-$5a   $61-$7a   $30-$39   

toupper:
	addlw	085h		; $c6-$df   $e6-$ff   $b5-$be
	addlw	01ah		; $e0-$f9   $00-$19   $cf-$d8
	btfsc	STATUS,C
	addlw	0e0h		;           $e0-$f9
	addlw	061h		; $41-$5a   $41-$7a   $30-$39
	return


;-----------------------------------------------------------------------------
; test whether character in W is a hex digit
;   returns with STATUS.Z clear if not hex
;   returns with STATUS.Z set if hex, and value in W
;-----------------------------------------------------------------------------

				; 0-9      A-F       a-f       Z
				; $30-39   $41-$46   $61-$66   $4a

ishex:	addlw	0c6h		; $f6-$ff  $07-$0c   $27-$2c   $10
	addlw	00ah		; $00-$09  $11-$16   $31-$36   $1a
	btfsc	STATUS,C
	goto	is09

	addlw	0e9h		;          $fa-$ff   $1a-$1f   $03
	addlw	006h		;          $00-$05   $20-$25   $09
	btfsc	STATUS,C
	goto	isaf

	addlw	0dah		;                    $fa-$ff   $e3
	addlw	006h		;                    $00-$05   $e9
	btfsc	STATUS,C
	goto	isaf

	addlw	061h		;                              $4a
	bcf	STATUS,Z
	return

isaf:	addlw	00ah
is09:	bsf	STATUS,Z
	return


;-----------------------------------------------------------------------------
; command scanner
;-----------------------------------------------------------------------------

; Enter with W-reg holding the command character and
; "CmdPtr" holding the address of the first byte of table to be scanned.
; The table must be terminated with a null "command" =0 which has its own
; address entry to handle the "command not found" condition.

CmdScan:
	movwf	temp		;Save the command
	movlw	cmdtbl_page
	movwf	PCLATH
Cmd10:	movf	cmdptr,W
	call	ScanGo
	andlw	0FFh		;Command table terminator?
	btfsc	STATUS,Z
	goto	Cmd20		; Yes, then do "no command found" and return
	xorwf	temp,W		;Does table entry match command?
	btfsc	STATUS,Z
	goto	Cmd20		; Yes, then do the command found
	incf	cmdptr		;Point to next command in table
	incf	cmdptr
	goto	Cmd10

Cmd20:	incf	cmdptr,W	;Execute the command and return from there
ScanGo:	movwf	PCL		;Get command from table


;-----------------------------------------------------------------------------
; commands
;-----------------------------------------------------------------------------

commacmd:
	movlw	','
	call	xmit
	movf	arg1,W
	movwf	arg2
	movf	arg1+1,W
	movwf	arg2+1
	clrf	arg1
	clrf	arg1+1
	bsf	cmdflag,gotarg2
	bcf	STATUS,Z
	return


loopcmd:
	call	echocmd
lc0:	call	recv
	movwf	temp
	xorlw	esc
	btfsc	STATUS,Z
	goto	lc9

	movf	temp,w
	call	xmit
	goto	lc0

lc9:	call	outcrlf
	goto	cmddone


numcmd:	call	echocmd

	btfsc	cmdflag,gotarg1
	goto	num1
	movlw	noargs_msg
	call	msgout
	goto	cmddone

num1:	btfss	cmdflag,gotarg2
	goto	num2

	movf	arg2+1,w
	call	outhb
	movf	arg2,w
	call	outhb
	movlw	','
	call	xmit

num2:	movf	arg1+1,w
	call	outhb
	movf	arg1,w
	call	outhb

	call	outcrlf
	goto	cmddone


badcmd:
;	movlw	badcmd_msg
;	call	msgout
	bcf	STATUS,Z
	return


echocmd:	
	movf	cmd,w
	call	xmit
outcrlf:
	movlw	crlf_msg
	goto	msgout


cmddone:
	bsf	STATUS,Z
	return


;-----------------------------------------------------------------------------
; command table
;-----------------------------------------------------------------------------

	org	cmdtbl_page<<8

cmdtbl:
	retlw	','
	goto	commacmd
	retlw	'L'
	goto	loopcmd
	retlw	'N'
	goto	numcmd
	retlw	cr
	goto	outcrlf
	retlw	0
	goto	badcmd


;-----------------------------------------------------------------------------
; tables
;-----------------------------------------------------------------------------

	org	message_page<<8

table:	addwf	PCL

banner	equ	$-(table+1)
	dt	cr,lf,"Hello, World!",cr,lf,0
prompt	equ	$-(table+1)
	dt	"Command? ",0
badcmd_msg	equ	$-(table+1)
	dt	"Unrecognized command",0
crlf_msg	equ	$-(table+1)
	dt	cr,lf,0
noargs_msg	equ	$-(table+1)
	dt	"No args",cr,lf,0

	end
